@@ -29,7 +29,6 @@
 #endif
 
 #include <unistd.h>
-#include <xf86_OSproc.h>
 #include <xf86Parser.h>
 #include <xf86drm.h>
 #include <xf86drmMode.h>
@@ -160,19 +159,19 @@ static const SymTabRec intel_chipsets[] = {
 	{PCI_CHIP_IVYBRIDGE_D_GT2,		"Ivybridge Desktop (GT2)" },
 	{PCI_CHIP_IVYBRIDGE_S_GT1,		"Ivybridge Server" },
 	{PCI_CHIP_IVYBRIDGE_S_GT2,		"Ivybridge Server (GT2)" },
-	{PCI_CHIP_HASWELL_D_GT1,		"Haswell Desktop (GT1)" },
-	{PCI_CHIP_HASWELL_D_GT2,		"Haswell Desktop (GT2)" },
+	{PCI_CHIP_HASWELL_D_GT1,		"HD Graphics" },
+	{PCI_CHIP_HASWELL_D_GT2,		"HD Graphics 4600" },
 	{PCI_CHIP_HASWELL_D_GT3,		"Haswell Desktop (GT3)" },
-	{PCI_CHIP_HASWELL_M_GT1,		"Haswell Mobile (GT1)" },
-	{PCI_CHIP_HASWELL_M_GT2,		"Haswell Mobile (GT2)" },
+	{PCI_CHIP_HASWELL_M_GT1,		"HD Graphics" },
+	{PCI_CHIP_HASWELL_M_GT2,		"HD Graphics 4600" },
 	{PCI_CHIP_HASWELL_M_GT3,		"Haswell Mobile (GT3)" },
-	{PCI_CHIP_HASWELL_S_GT1,		"Haswell Server (GT1)" },
-	{PCI_CHIP_HASWELL_S_GT2,		"Haswell Server (GT2)" },
+	{PCI_CHIP_HASWELL_S_GT1,		"HD Graphics" },
+	{PCI_CHIP_HASWELL_S_GT2,		"HD Graphics P4600/P4700" },
 	{PCI_CHIP_HASWELL_S_GT3,		"Haswell Server (GT3)" },
 	{PCI_CHIP_HASWELL_B_GT1,		"Haswell (GT1)" },
 	{PCI_CHIP_HASWELL_B_GT2,		"Haswell (GT2)" },
 	{PCI_CHIP_HASWELL_B_GT3,		"Haswell (GT3)" },
-	{PCI_CHIP_HASWELL_E_GT1,		"Haswell (GT1)" },
+	{PCI_CHIP_HASWELL_E_GT1,		"HD Graphics" },
 	{PCI_CHIP_HASWELL_E_GT2,		"Haswell (GT2)" },
 	{PCI_CHIP_HASWELL_E_GT3,		"Haswell (GT3)" },
 	{PCI_CHIP_HASWELL_SDV_D_GT1,		"Haswell SDV Desktop (GT1)" },
@@ -192,34 +191,34 @@ static const SymTabRec intel_chipsets[] = {
 	{PCI_CHIP_HASWELL_SDV_E_GT3,		"Haswell SDV (GT3)" },
 	{PCI_CHIP_HASWELL_ULT_D_GT1,		"Haswell ULT Desktop (GT1)" },
 	{PCI_CHIP_HASWELL_ULT_D_GT2,		"Haswell ULT Desktop (GT2)" },
-	{PCI_CHIP_HASWELL_ULT_D_GT3,		"Haswell ULT Desktop (GT3)" },
-	{PCI_CHIP_HASWELL_ULT_M_GT1,		"Haswell ULT Mobile (GT1)" },
-	{PCI_CHIP_HASWELL_ULT_M_GT2,		"Haswell ULT Mobile (GT2)" },
-	{PCI_CHIP_HASWELL_ULT_M_GT3,		"Haswell ULT Mobile (GT3)" },
+	{PCI_CHIP_HASWELL_ULT_D_GT3,		"Iris(TM) Graphics 5100" },
+	{PCI_CHIP_HASWELL_ULT_M_GT1,		"HD Graphics" },
+	{PCI_CHIP_HASWELL_ULT_M_GT2,		"HD Graphics 4400" },
+	{PCI_CHIP_HASWELL_ULT_M_GT3,		"HD Graphics 5000" },
 	{PCI_CHIP_HASWELL_ULT_S_GT1,		"Haswell ULT Server (GT1)" },
 	{PCI_CHIP_HASWELL_ULT_S_GT2,		"Haswell ULT Server (GT2)" },
-	{PCI_CHIP_HASWELL_ULT_S_GT3,		"Haswell ULT Server (GT3)" },
+	{PCI_CHIP_HASWELL_ULT_S_GT3,		"Iris(TM) Graphics 5100" },
 	{PCI_CHIP_HASWELL_ULT_B_GT1,		"Haswell ULT (GT1)" },
 	{PCI_CHIP_HASWELL_ULT_B_GT2,		"Haswell ULT (GT2)" },
-	{PCI_CHIP_HASWELL_ULT_B_GT3,		"Haswell ULT (GT3)" },
-	{PCI_CHIP_HASWELL_ULT_E_GT1,		"Haswell ULT (GT1)" },
-	{PCI_CHIP_HASWELL_ULT_E_GT2,		"Haswell ULT (GT2)" },
-	{PCI_CHIP_HASWELL_ULT_E_GT3,		"Haswell ULT (GT3)" },
+	{PCI_CHIP_HASWELL_ULT_B_GT3,		"Iris(TM) Graphics 5100" },
+	{PCI_CHIP_HASWELL_ULT_E_GT1,		"HD Graphics" },
+	{PCI_CHIP_HASWELL_ULT_E_GT2,		"HD Graphics 4200" },
+	{PCI_CHIP_HASWELL_ULT_E_GT3,		"Iris(TM) Graphics 5100" },
 	{PCI_CHIP_HASWELL_CRW_D_GT1,		"Haswell CRW Desktop (GT1)" },
-	{PCI_CHIP_HASWELL_CRW_D_GT2,		"Haswell CRW Desktop (GT2)" },
-	{PCI_CHIP_HASWELL_CRW_D_GT3,		"Haswell CRW Desktop (GT3)" },
+	{PCI_CHIP_HASWELL_CRW_D_GT2,		"HD Graphics 4600" },
+	{PCI_CHIP_HASWELL_CRW_D_GT3,		"Iris(TM) Pro Graphics 5200" },
 	{PCI_CHIP_HASWELL_CRW_M_GT1,		"Haswell CRW Mobile (GT1)" },
-	{PCI_CHIP_HASWELL_CRW_M_GT2,		"Haswell CRW Mobile (GT2)" },
-	{PCI_CHIP_HASWELL_CRW_M_GT3,		"Haswell CRW Mobile (GT3)" },
+	{PCI_CHIP_HASWELL_CRW_M_GT2,		"HD Graphics 4600" },
+	{PCI_CHIP_HASWELL_CRW_M_GT3,		"Iris(TM) Pro Graphics 5200" },
 	{PCI_CHIP_HASWELL_CRW_S_GT1,		"Haswell CRW Server (GT1)" },
 	{PCI_CHIP_HASWELL_CRW_S_GT2,		"Haswell CRW Server (GT2)" },
-	{PCI_CHIP_HASWELL_CRW_S_GT3,		"Haswell CRW Server (GT3)" },
+	{PCI_CHIP_HASWELL_CRW_S_GT3,		"Iris(TM) Pro Graphics 5200" },
 	{PCI_CHIP_HASWELL_CRW_B_GT1,		"Haswell CRW (GT1)" },
 	{PCI_CHIP_HASWELL_CRW_B_GT2,		"Haswell CRW (GT2)" },
-	{PCI_CHIP_HASWELL_CRW_B_GT3,		"Haswell CRW (GT3)" },
+	{PCI_CHIP_HASWELL_CRW_B_GT3,		"Iris(TM) Pro Graphics 5200" },
 	{PCI_CHIP_HASWELL_CRW_E_GT1,		"Haswell CRW (GT1)" },
 	{PCI_CHIP_HASWELL_CRW_E_GT2,		"Haswell CRW (GT2)" },
-	{PCI_CHIP_HASWELL_CRW_E_GT3,		"Haswell CRW (GT3)" },
+	{PCI_CHIP_HASWELL_CRW_E_GT3,		"Iris(TM) Pro Graphics 5200" },
 	{PCI_CHIP_VALLEYVIEW_PO,		"ValleyView PO board" },
 	{-1,					NULL}
 };
@@ -435,49 +434,46 @@ static Bool intel_driver_func(ScrnInfoPtr pScrn,
 	}
 }
 
-static Bool has_kernel_mode_setting(const struct pci_device *dev)
+static Bool is_i915_device(int fd)
+{
+	drm_version_t version;
+	char name[5] = "";
+
+	memset(&version, 0, sizeof(version));
+	version.name_len = 4;
+	version.name = name;
+
+	if (drmIoctl(fd, DRM_IOCTL_VERSION, &version))
+		return FALSE;
+
+	return strcmp("i915", name) == 0;
+}
+
+static Bool has_kernel_mode_setting(int entity_num,
+				    const struct pci_device *dev,
+				    const char *path)
 {
-	char id[20];
 	int ret, fd;
 
-	snprintf(id, sizeof(id),
-		 "pci:%04x:%02x:%02x.%d",
-		 dev->domain, dev->bus, dev->dev, dev->func);
-
-	ret = drmCheckModesettingSupported(id);
-	if (ret) {
-		if (xf86LoadKernelModule("i915"))
-			ret = drmCheckModesettingSupported(id);
-		if (ret)
-			return FALSE;
-		/* Be nice to the user and load fbcon too */
-		(void)xf86LoadKernelModule("fbcon");
-	}
+	fd = intel_open_device(entity_num, dev, path);
+	if (fd == -1)
+		return FALSE;
 
 	/* Confirm that this is a i915.ko device with GEM/KMS enabled */
-	ret = FALSE;
-	fd = drmOpen(NULL, id);
-	if (fd != -1) {
-		drmVersionPtr version = drmGetVersion(fd);
-		if (version) {
-			ret = strcmp ("i915", version->name) == 0;
-			drmFreeVersion(version);
-		}
-		if (ret) {
-			struct drm_i915_getparam gp;
-			gp.param = I915_PARAM_HAS_GEM;
-			gp.value = &ret;
-			if (drmIoctl(fd, DRM_IOCTL_I915_GETPARAM, &gp))
-				ret = FALSE;
-		}
-		if (ret) {
-			struct drm_mode_card_res res;
+	ret = is_i915_device(fd);
+	if (ret) {
+		struct drm_i915_getparam gp;
+		gp.param = I915_PARAM_HAS_GEM;
+		gp.value = &ret;
+		if (drmIoctl(fd, DRM_IOCTL_I915_GETPARAM, &gp))
+			ret = FALSE;
+	}
+	if (ret) {
+		struct drm_mode_card_res res;
 
-			memset(&res, 0, sizeof(res));
-			if (drmIoctl(fd, DRM_IOCTL_MODE_GETRESOURCES, &res))
-				ret = FALSE;
-		}
-		close(fd);
+		memset(&res, 0, sizeof(res));
+		if (drmIoctl(fd, DRM_IOCTL_MODE_GETRESOURCES, &res))
+			ret = FALSE;
 	}
 
 	return ret;
@@ -578,7 +574,7 @@ static Bool intel_pci_probe(DriverPtr		driver,
 			    struct pci_device	*device,
 			    intptr_t		match_data)
 {
-	if (!has_kernel_mode_setting(device)) {
+	if (!has_kernel_mode_setting(entity_num, device, NULL)) {
 #if KMS_ONLY
 		return FALSE;
 #else
@@ -609,7 +605,8 @@ intel_platform_probe(DriverPtr driver,
 	if (!dev->pdev)
 		return FALSE;
 
-	if (!has_kernel_mode_setting(dev->pdev))
+	if (!has_kernel_mode_setting(entity_num, dev->pdev,
+				     xf86_get_platform_device_attrib(dev, ODEV_ATTRIB_PATH)))
 		return FALSE;
 
 	/* Allow ourselves to act as a slaved output if not primary */
