@@ -39,6 +39,7 @@
 #define pure __attribute__((pure))
 #define __packed__ __attribute__((__packed__))
 #define flatten __attribute__((flatten))
+#define page_aligned __attribute__((aligned(4096)))
 #else
 #define likely(expr) (expr)
 #define unlikely(expr) (expr)
@@ -50,6 +51,7 @@
 #define pure
 #define __packed__
 #define flatten
+#define page_aligned
 #endif
 
 #define HAS_GCC(major, minor) defined(__GNUC__) && (__GNUC__ > (major) || __GNUC__ == (major) && __GNUC_MINOR__ >= (minor))
@@ -63,6 +65,12 @@
 #define avx2 __attribute__((target("avx2,sse4.2,sse2,fpmath=sse")))
 #endif
 
+#if HAS_GCC(4, 5) && defined(__OPTIMIZE__)
+#define fast_memcpy __attribute__((optimize("Ofast"))) __attribute__((target("inline-all-stringops")))
+#else
+#define fast_memcpy
+#endif
+
 #ifdef HAVE_VALGRIND
 #define VG(x) x
 #else
